"""Methods for command-line parsing of a Rhasspy Hermes client."""
import argparse
import logging
import os
import ssl

import paho.mqtt.client as mqtt


def add_hermes_args(parser: argparse.ArgumentParser):
    """Add shared Hermes/MQTT command-line arguments.

    These are useful arguments for every Hermes client, concerning the connection,
    authentication, site IDs, debugging and logging.
    """
    parser.add_argument(
        "--host", default="localhost", help="MQTT host (default: localhost)"
    )
    parser.add_argument(
        "--port", type=int, default=1883, help="MQTT port (default: 1883)"
    )
    parser.add_argument("--username", help="MQTT username")
    parser.add_argument("--password", help="MQTT password")

    parser.add_argument("--tls", action="store_true", help="Enable MQTT TLS")
    parser.add_argument(
        "--tls-ca-certs", help="MQTT TLS Certificate Authority certificate files"
    )
    parser.add_argument("--tls-certfile", help="MQTT TLS client certificate file (PEM)")
    parser.add_argument("--tls-keyfile", help="MQTT TLS client key file (PEM)")
    parser.add_argument(
        "--tls-cert-reqs",
        default="CERT_REQUIRED",
        choices=["CERT_REQUIRED", "CERT_OPTIONAL", "CERT_NONE"],
        help="MQTT TLS certificate requirements for broker (default: CERT_REQUIRED)",
    )
    parser.add_argument(
        "--tls-version", type=int, help="MQTT TLS version (default: highest)"
    )
    parser.add_argument("--tls-ciphers", help="MQTT TLS ciphers to use")

    parser.add_argument(
        "--site-id",
        action="append",
        help="Hermes site id(s) to listen for (default: all)",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG messages to the console"
    )
    parser.add_argument(
        "--log-format",
        default="[%(levelname)s:%(asctime)s] %(name)s: %(message)s",
        help="Python logger format",
    )


def setup_logging(args: argparse.Namespace):
    """Set up Python logging."""
    if args.debug:
        logging.basicConfig(level=logging.DEBUG, format=args.log_format)
    else:
        logging.basicConfig(level=logging.INFO, format=args.log_format)


def connect(client: mqtt.Client, args: argparse.Namespace):
    """Connect to an MQTT broker with supplied arguments."""
    if args.username:
        client.username_pw_set(args.username, args.password)

    # TLS
    if args.tls:
        # TLS is enabled
        if args.tls_version is None:
            # Use highest TLS version
            args.tls_version = ssl.PROTOCOL_TLS

        if args.tls_certfile is not None:
            args.tls_certfile = os.path.expandvars(args.tls_certfile)
        if args.tls_keyfile is not None:
            args.tls_keyfile = os.path.expandvars(args.tls_keyfile)

        client.tls_set(
            ca_certs=args.tls_ca_certs,
            certfile=args.tls_certfile,
            keyfile=args.tls_keyfile,
            cert_reqs=getattr(ssl, args.tls_cert_reqs),
            tls_version=args.tls_version,
            ciphers=(args.tls_ciphers or None),
        )

    client.connect(args.host, args.port)



"""Tests for rhasspyhermes.audioserver"""
from rhasspyhermes.audioserver import AudioFrame, AudioPlayBytes, AudioPlayFinished

site_id = "testSiteId"
request_id = "testRequestId"


def test_audio_frame():
    """Test AudioFrame."""
    assert AudioFrame.is_topic(AudioFrame.topic(site_id=site_id))
    assert AudioFrame.get_site_id(AudioFrame.topic(site_id=site_id)) == site_id


def test_audio_play_bytes():
    """Test AudioPlayBytes."""
    assert AudioPlayBytes.is_topic(
        AudioPlayBytes.topic(site_id=site_id, request_id=request_id)
    )
    assert (
        AudioPlayBytes.get_site_id(
            AudioPlayBytes.topic(site_id=site_id, request_id=request_id)
        )
        == site_id
    )
    assert (
        AudioPlayBytes.get_request_id(
            AudioPlayBytes.topic(site_id=site_id, request_id=request_id)
        )
        == request_id
    )


def test_audio_play_finished():
    """Test AudioPlayFinished."""
    assert AudioPlayFinished.is_topic(AudioPlayFinished.topic(site_id=site_id))
    assert (
        AudioPlayFinished.get_site_id(AudioPlayFinished.topic(site_id=site_id))
        == site_id
    )



"""Rhasspy-only messages for intent handling."""
from dataclasses import dataclass

from .base import Message


@dataclass
class HandleToggleOn(Message):
    """Enable intent handling.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/handle/toggleOn``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - siteId
            - String
            - The id of the site where intent handling should be enabled. Defaults to ``"default"``.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'rhasspy/handle/toggleOn' -m '{"siteId": "default"}'

    Example
    -------

    >>> from rhasspyhermes.handle import HandleToggleOn
    >>> on = HandleToggleOn()
    >>> on
    HandleToggleOn(site_id='default')
    >>> on.payload()
    '{"siteId": "default"}'
    >>> on.topic()
    'rhasspy/handle/toggleOn'

    Note
    ----

    This is a Rhasspy-only message."""

    site_id: str = "default"
    """The id of the site where intent handling should be enabled"""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"rhasspy/handle/toggleOn"``
        """
        return "rhasspy/handle/toggleOn"


@dataclass
class HandleToggleOff(Message):
    """Disable intent handling.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/handle/toggleOff``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - siteId
            - String
            - The id of the site where intent handling should be disabled. Defaults to ``"default"``.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'rhasspy/handle/toggleOff' -m '{"siteId": "default"}'

    Example
    -------

    >>> from rhasspyhermes.handle import HandleToggleOff
    >>> off = HandleToggleOff()
    >>> off
    HandleToggleOff(site_id='default')
    >>> off.payload()
    '{"siteId": "default"}'
    >>> off.topic()
    'rhasspy/handle/toggleOff'

    Note
    ----

    This is a Rhasspy-only message."""

    site_id: str = "default"
    """The id of the site where intent handling should be disabled"""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"rhasspy/handle/toggleOff"``
        """
        return "rhasspy/handle/toggleOff"



"""Messages for audio recording and playback."""
import audioop
import io
import re
import time
import typing
import wave
from dataclasses import dataclass
from enum import Enum

from .base import Message


@dataclass
class AudioFrame(Message):
    """Recorded frame of audio.

    Attributes
    ----------
    wav_bytes: bytes
        Recorded audio frame in WAV format
    """

    TOPIC_PATTERN = re.compile(r"^hermes/audioServer/([^/]+)/audioFrame$")

    wav_bytes: bytes

    def payload(self) -> typing.Union[str, bytes]:
        """Get binary/string for this message."""
        return self.wav_bytes

    @classmethod
    def is_binary_payload(cls) -> bool:
        """True if payload is not JSON."""
        return True

    @classmethod
    def is_site_in_topic(cls) -> bool:
        """True if site id is in topic."""
        return True

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get topic for message."""
        site_id = kwargs.get("site_id", "+")
        return f"hermes/audioServer/{site_id}/audioFrame"

    @classmethod
    def get_site_id(cls, topic: str) -> typing.Optional[str]:
        """Get site id from a topic"""
        match = re.match(AudioFrame.TOPIC_PATTERN, topic)
        assert match, "Not an audioFrame topic"
        return match.group(1)

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template"""
        return re.match(AudioFrame.TOPIC_PATTERN, topic) is not None

    @classmethod
    def iter_wav_chunked(
        cls, wav_io: typing.BinaryIO, frames_per_chunk: int, live_delay: bool = False
    ) -> typing.Iterable[bytes]:
        """Split single WAV into multiple WAV chunks"""
        with wave.open(wav_io) as in_wav:
            frames_left = in_wav.getnframes()

            while frames_left > 0:
                chunk = in_wav.readframes(frames_per_chunk)
                if not chunk:
                    break

                # Wrap chunk in WAV
                with io.BytesIO() as out_io:
                    out_wav: wave.Wave_write = wave.open(out_io, "wb")
                    with out_wav:
                        out_wav.setframerate(in_wav.getframerate())
                        out_wav.setsampwidth(in_wav.getsampwidth())
                        out_wav.setnchannels(in_wav.getnchannels())
                        out_wav.writeframes(chunk)

                    wav_bytes = out_io.getvalue()
                    yield wav_bytes

                    if live_delay:
                        time.sleep(AudioFrame.get_wav_duration(wav_bytes))

                frames_left -= frames_per_chunk

    @classmethod
    def get_wav_duration(cls, wav_bytes: bytes) -> float:
        """Return the real-time duration of a WAV file"""
        with io.BytesIO(wav_bytes) as wav_buffer:
            wav_file: wave.Wave_read = wave.open(wav_buffer, "rb")
            with wav_file:
                frames = wav_file.getnframes()
                rate = wav_file.getframerate()
                return frames / float(rate)


@dataclass
class AudioPlayBytes(Message):
    """Play WAV sound on specific site.

    Attributes
    ----------
    wav_bytes: bytes
        Audio to play in WAV format
    """

    TOPIC_PATTERN = re.compile(r"^hermes/audioServer/([^/]+)/playBytes/([^/]+)$")

    wav_bytes: bytes

    def payload(self) -> typing.Union[str, bytes]:
        """Get binary/string for this message."""
        return self.wav_bytes

    @classmethod
    def is_binary_payload(cls) -> bool:
        """True if payload is not JSON."""
        return True

    @classmethod
    def is_site_in_topic(cls) -> bool:
        """True if site id is in topic."""
        return True

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get topic for message."""
        site_id = kwargs.get("site_id", "+")
        request_id = kwargs.get("request_id", "#")
        return f"hermes/audioServer/{site_id}/playBytes/{request_id}"

    @classmethod
    def get_site_id(cls, topic: str) -> typing.Optional[str]:
        """Get site id from a topic"""
        match = re.match(AudioPlayBytes.TOPIC_PATTERN, topic)
        assert match, "Not a playBytes topic"
        return match.group(1)

    @classmethod
    def get_request_id(cls, topic: str) -> str:
        """Get request id from a topic"""
        match = re.match(AudioPlayBytes.TOPIC_PATTERN, topic)
        assert match, "Not a playBytes topic"
        return match.group(2)

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template"""
        return re.match(AudioPlayBytes.TOPIC_PATTERN, topic) is not None


@dataclass
class AudioPlayFinished(Message):
    """Sent when audio service has finished playing a sound.

    Attributes
    ----------
    id: str : Optional[str] = None
        Request identifier for the request passed from playBytes topic

    session_id: Optional[str] = None
        The id of the session, if there is an active session
    """

    TOPIC_PATTERN = re.compile(r"^hermes/audioServer/([^/]+)/playFinished$")

    id: typing.Optional[str] = None
    session_id: typing.Optional[str] = None

    @classmethod
    def is_site_in_topic(cls) -> bool:
        """True if site id is in topic."""
        return True

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get topic for message."""
        site_id = kwargs.get("site_id", "+")
        return f"hermes/audioServer/{site_id}/playFinished"

    @classmethod
    def get_site_id(cls, topic: str) -> typing.Optional[str]:
        """Get site_id from a topic"""
        match = re.match(AudioPlayFinished.TOPIC_PATTERN, topic)
        assert match, "Not a playFinished topic"
        return match.group(1)

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template"""
        return re.match(AudioPlayFinished.TOPIC_PATTERN, topic) is not None


# -----------------------------------------------------------------------------
# Rhasspy Only
# -----------------------------------------------------------------------------


class AudioDeviceMode(str, Enum):
    """Mode of an audio device.

    Values
    ------
    INPUT
        Recording device

    OUTPUT
        Playback device
    """

    INPUT = "input"
    OUTPUT = "output"


@dataclass
class AudioDevice:
    """Description of an audio device.

    Attributes
    ----------
    mode: AudioDeviceMode
        Recording or playback device

    id: str
        Unique id of audio device

    name: Optional[str] = None
        Optional human-readable name of audio device

    description: Optional[str] = None
        Optional human-readable description of audio device

    working: Optional[bool] = None
        Status of audio device if tested
    """

    mode: AudioDeviceMode
    id: str
    name: typing.Optional[str] = None
    description: typing.Optional[str] = None
    working: typing.Optional[bool] = None


@dataclass
class AudioGetDevices(Message):
    """Get details for available audio devices.

    Attributes
    ----------
    modes: List[AudioDeviceMode]
        Device types to get information about

    id: Optional[str] = None
        Unique id to be returned in response

    site_id: str = "default"
        Id of the site where devices are located

    test: bool = False
        True if devices should be tested
    """

    modes: typing.List[AudioDeviceMode]
    site_id: str = "default"
    id: typing.Optional[str] = None
    test: bool = False

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get topic for message."""
        return "rhasspy/audioServer/getDevices"


@dataclass
class AudioDevices(Message):
    """Response to getDevices.

    Attributes
    ----------
    devices: List[AudioDevice]
        Description of requested device types

    id: Optional[str] = None
        Unique id from request

    site_id: str = "default"
        Id of site where devices are located
    """

    devices: typing.List[AudioDevice]
    site_id: str = "default"
    id: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get topic for message."""
        return "rhasspy/audioServer/devices"


@dataclass
class AudioSessionFrame(Message):
    """Recorded audio frame for a specific session.

    Attributes
    ----------
    wav_bytes: bytes
        Audio frame in WAV format
    """

    TOPIC_PATTERN = re.compile(
        r"^hermes/audioServer/([^/]+)/([^/]+)/audioSessionFrame$"
    )

    wav_bytes: bytes

    def payload(self) -> typing.Union[str, bytes]:
        """Get binary/string for this message."""
        return self.wav_bytes

    @classmethod
    def is_binary_payload(cls) -> bool:
        """True if payload is not JSON."""
        return True

    @classmethod
    def is_site_in_topic(cls) -> bool:
        """True if site id is in topic."""
        return True

    @classmethod
    def is_session_in_topic(cls) -> bool:
        """True if session id is in topic."""
        return True

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get topic for message."""
        site_id = kwargs.get("site_id", "+")
        session_id = kwargs.get("session_id", "+")
        return f"hermes/audioServer/{site_id}/{session_id}/audioSessionFrame"

    @classmethod
    def get_site_id(cls, topic: str) -> typing.Optional[str]:
        """Get site id from a topic"""
        match = re.match(AudioSessionFrame.TOPIC_PATTERN, topic)
        assert match, "Not an audioSessionFrame topic"
        return match.group(1)

    @classmethod
    def get_session_id(cls, topic: str) -> typing.Optional[str]:
        """Get session id from a topic"""
        match = re.match(AudioSessionFrame.TOPIC_PATTERN, topic)
        assert match, "Not an audioSessionFrame topic"
        return match.group(2)

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template"""
        return re.match(AudioSessionFrame.TOPIC_PATTERN, topic) is not None


@dataclass
class AudioSummary(Message):
    """Summary of recent audio frame(s) for diagnostic purposes.

    debiased_energy: float
        Audio energy computed using get_debiased_energy

    is_speech: typing.Optional[bool] = None
        True/false if VAD detected speech
    """

    TOPIC_PATTERN = re.compile(r"^hermes/audioServer/([^/]+)/audioSummary$")

    debiased_energy: float
    is_speech: typing.Optional[bool] = None

    @classmethod
    def get_debiased_energy(cls, audio_data: bytes) -> float:
        """Compute RMS of debiased audio."""
        # Thanks to the speech_recognition library!
        # https://github.com/Uberi/speech_recognition/blob/master/speech_recognition/__init__.py
        energy = -audioop.rms(audio_data, 2)
        energy_bytes = bytes([energy & 0xFF, (energy >> 8) & 0xFF])
        debiased_energy = audioop.rms(
            audioop.add(audio_data, energy_bytes * (len(audio_data) // 2), 2), 2
        )

        # Probably actually audio if > 30
        return debiased_energy

    @classmethod
    def is_site_in_topic(cls) -> bool:
        """True if site id is in topic."""
        return True

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get topic for message."""
        site_id = kwargs.get("site_id", "+")
        return f"hermes/audioServer/{site_id}/audioSummary"

    @classmethod
    def get_site_id(cls, topic: str) -> typing.Optional[str]:
        """Get site id from a topic"""
        match = re.match(AudioSummary.TOPIC_PATTERN, topic)
        assert match, "Not an audioSummary topic"
        return match.group(1)

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template"""
        return re.match(AudioSummary.TOPIC_PATTERN, topic) is not None


@dataclass
class SummaryToggleOn(Message):
    """Activate sending of audio summaries.

    Attributes
    ----------
    site_id: str = "default"
        Id of site where audio is being recorded
    """

    site_id: str = "default"

    @classmethod
    def topic(cls, **kwargs) -> str:
        return "hermes/audioServer/toggleSummaryOn"


@dataclass
class SummaryToggleOff(Message):
    """Deactivate sending of audio summaries.

    Attributes
    ----------
    site_id: str = "default"
        Id of site where audio is being recorded
    """

    site_id: str = "default"

    @classmethod
    def topic(cls, **kwargs) -> str:
        return "hermes/audioServer/toggleSummaryOff"


@dataclass
class AudioToggleOn(Message):
    """Activate audio output system.

    Attributes
    ----------
    site_id: str = "default"
        Id of site where audio should be turned off
    """

    site_id: str = "default"

    @classmethod
    def topic(cls, **kwargs) -> str:
        return "hermes/audioServer/toggleOn"


@dataclass
class AudioToggleOff(Message):
    """Deactivate audio output system.

    Attributes
    ----------
    site_id: str = "default"
        Id of site where audio should be turned on
    """

    site_id: str = "default"

    @classmethod
    def topic(cls, **kwargs) -> str:
        return "hermes/audioServer/toggleOff"


@dataclass
class AudioRecordError(Message):
    """Error from audio input component.

    Attributes
    ----------
    error: str
        A description of the error that occurred

    site_id: str = "default"
        The id of the site where the error occurred

    context: Optional[str] = None
        Additional information on the context in which the error occurred

    session_id: Optional[str] = None
        The id of the session, if there is an active session
    """

    error: str
    site_id: str = "default"
    context: typing.Optional[str] = None
    session_id: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/error/audioServer/record"


@dataclass
class AudioPlayError(Message):
    """Error from audio output component.

    Attributes
    ----------
    error: str
        A description of the error that occurred

    site_id: str = "default"
        The id of the site where the error occurred

    context: Optional[str] = None
        Additional information on the context in which the error occurred

    session_id: Optional[str] = None
        The id of the session, if there is an active session
    """

    error: str
    site_id: str = "default"
    context: typing.Optional[str] = None
    session_id: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/error/audioServer/play"



"""Rhasspy-only messages for intent training."""
import re
from dataclasses import dataclass

from .base import Message


@dataclass
class IntentGraphRequest(Message):
    """Request publication of intent graph from training.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/train/getIntentGraph``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - id
            - String
            - Unique id for request. Appended to reply topic (:class:`IntentGraph`).
          * - siteId
            - String
            - The id of the site where training occurred. Defaults to ``"default"``.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'rhasspy/train/getIntentGraph' -m '{"id": "abcd", "siteId": "default"}'

    Example
    -------

    >>> from rhasspyhermes.train import IntentGraphRequest
    >>> request = IntentGraphRequest(id='abcd')
    >>> request.payload()
    '{"id": "abcd", "siteId": "default"}'
    >>> request.topic()
    'rhasspy/train/getIntentGraph'

    Note
    ----

    This is a Rhasspy-only message."""

    id: str
    """Unique id for request. Appended to reply topic (:class:`IntentGraph`)."""
    site_id: str = "default"
    """The id of the site where training occurred."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"rhasspy/train/getIntentGraph"``
        """
        return "rhasspy/train/getIntentGraph"


@dataclass
class IntentGraph(Message):
    """Intent graph from training.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/train/intentGraph``

      Payload (binary)
        gzipped pickle bytes containing a NetworkX_ intent graph

      .. _NetworkX: https://networkx.github.io/

      Subscribe to this message type with ``mosquitto_sub`` and show the binary payload as hexadecimal numbers:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -t 'rhasspy/train/intentGraph' -F %x

    Note
    ----

    This is a Rhasspy-only message."""

    TOPIC_PATTERN = re.compile(r"^rhasspy/train/intentGraph/([^/]+)$")

    graph_bytes: bytes
    """Gzipped pickle bytes containing a NetworkX intent graph"""

    def payload(self) -> bytes:
        """Get the binary payload for this message.

        Returns
        -------

        bytes
            The binary payload as gzipped pickle bytes containing a NetworkX intent graph.
        """
        return self.graph_bytes

    @classmethod
    def is_binary_payload(cls) -> bool:
        """Check for binary payload of message.

        Returns
        -------
        bool
            ``True``
        """
        return True

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Arguments
        ---------
        request_id: str
            Unique id for request. Supplied in request topic (:class:`IntentGraphRequest`)

        Returns
        -------
        str
            ``"rhasspy/train/intentGraph/{request_id}"``

        Example
        -------

        >>> from rhasspyhermes.train import IntentGraph
        >>> IntentGraph.topic(request_id="abcd")
        'rhasspy/train/intentGraph/abcd'
        """
        request_id = kwargs.get("request_id", "#")
        return f"rhasspy/train/intentGraph/{request_id}"

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """Check whether topic is for this message type.

        Arguments
        ---------
        topic
            message topic

        Returns
        -------
        bool
            ``True`` if topic is for this message type

        Example
        -------

        >>> from rhasspyhermes.train import IntentGraph
        >>> IntentGraph.is_topic("rhasspy/train/intentGraph/abcd")
        True
        """
        return re.match(IntentGraph.TOPIC_PATTERN, topic) is not None



"""Support for the Rhasspy Hermes protocol.

The Rhasspy Hermes protocol is an extension of the Snips Hermes protocol.
"""
import typing
from abc import ABCMeta

from dataclasses_json import DataClassJsonMixin, LetterCase, dataclass_json


@dataclass_json(letter_case=LetterCase.CAMEL)
class Message(DataClassJsonMixin, metaclass=ABCMeta):
    """Base class for Hermes messages.

    All classes implementing Hermes messages are subclasses of this class."""

    def __init__(self, **kwargs):
        DataClassJsonMixin.__init__(self, letter_case=LetterCase.CAMEL)

    def payload(self) -> typing.Union[str, bytes]:
        """Get the payload for this message.

        Returns
        -------

        Union[str, bytes]
            The payload as a JSON string or bytes

        Example
        -------

        >>> from rhasspyhermes.handle import HandleToggleOn
        >>> on = HandleToggleOn(site_id='satellite')
        >>> on.payload()
        '{"siteId": "satellite"}'
        """
        return self.to_json()

    @classmethod
    def get_site_id(cls, topic: str) -> typing.Optional[str]:
        """Extract site id from message topic.

        Arguments
        ---------
        topic
            message topic

        Returns
        -------
        Optional[str]
            The optional site id for this message topic

        Example
        -------

        >>> from rhasspyhermes.audioserver import AudioSessionFrame
        >>> topic = "hermes/audioServer/satellite/abcd/audioSessionFrame"
        >>> AudioSessionFrame.get_site_id(topic)
        'satellite'
        """
        return None

    @classmethod
    def get_session_id(cls, topic: str) -> typing.Optional[str]:
        """Extract session id from message topic.

        Arguments
        ---------
        topic
            message topic

        Returns
        -------
        Optional[str]
            The optional session id for this message topic

        Example
        -------

        >>> from rhasspyhermes.audioserver import AudioSessionFrame
        >>> topic = "hermes/audioServer/satellite/abcd/audioSessionFrame"
        >>> AudioSessionFrame.get_session_id(topic)
        'abcd'
        """
        return None

    @classmethod
    def is_binary_payload(cls) -> bool:
        """Check for binary payload of message.

        Returns
        -------
        bool
            ``True`` if message payload is not JSON

        Example
        -------

        >>> from rhasspyhermes.audioserver import AudioFrame
        >>> AudioFrame.is_binary_payload()
        True
        """
        return False

    @classmethod
    def is_site_in_topic(cls) -> bool:
        """Check for site id in topic.

        Returns
        -------
        bool
            ``True`` if site id is part of topic

        Example
        -------

        >>> from rhasspyhermes.asr import AsrTrain
        >>> AsrTrain.is_site_in_topic()
        True
        """
        return False

    @classmethod
    def is_session_in_topic(cls) -> bool:
        """Check for session id in topic.

        Returns
        -------
        bool
            ``True`` if session id is part of topic

        Example
        -------

        >>> from rhasspyhermes.asr import AsrAudioCaptured
        >>> AsrAudioCaptured.is_session_in_topic()
        True
        """
        return False

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            The MQTT topic for this message type

        Example
        -------

        >>> from rhasspyhermes.nlu import NluIntent
        >>> NluIntent.topic()
        'hermes/intent/#'
        """

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """Check whether topic is for this message type.

        Arguments
        ---------
        topic
            message topic

        Returns
        -------
        bool
            ``True`` if topic is for this message type

        Example
        --------

        >>> from rhasspyhermes.wake import HotwordDetected
        >>> HotwordDetected.is_topic("hermes/hotword/precise/detected")
        True
        """
        return topic == cls.topic()



"""Messages for natural language understanding."""
import re
import typing
from dataclasses import dataclass

from dataclasses_json import LetterCase, dataclass_json

from .base import Message
from .intent import Intent, Slot


@dataclass
class NluQuery(Message):
    """Request intent recognition from NLU component.

    .. admonition:: MQTT message

      Topic
        ``hermes/nlu/query``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - input
            - String
            - The text to send to the NLU component.
          * - siteId
            - String
            - The id of the site where the NLU component is located. Defaults to ``"default"``.
          * - id
            - String (optional)
            - A request identifier. If provided, it will be passed back in the response
              (:class:`NluIntentParsed` or :class:`NluIntentNotRecognized`).
          * - intentFilter
            - List of strings (optional)
            - A list of intent names to restrict the NLU resolution on.
          * - sessionId
            - String (optional)
            - The id of the session, if there is an active session.
          * - wakewordId
            - String (optional)
            - The id of the wakeword used to activate the ASR.
          * - lang
            - String (optional)
            - The language of the session.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'hermes/nlu/query' -m '{"input": "what time is it", "siteId": "default"}'

    Example
    -------

    >>> from rhasspyhermes.nlu import NluQuery
    >>> query = NluQuery(input='what time is it')
    >>> query.payload()
    '{"input": "what time is it", "siteId": "default", "id": null, "intentFilter": null, "sessionId": null, "wakewordId": null, "lang": null}'
    >>> query.topic()
    'hermes/nlu/query'
    """

    input: str
    """The text to send to the NLU component."""
    site_id: str = "default"
    """The id of the site where the NLU component is located.

    Note
    ----

    In contrast to the Snips Hermes protocol, the site id is compulsory.
    """
    id: typing.Optional[str] = None
    """A request identifier. If provided, it will be passed back in the response
    (:class:`NluIntentParsed` or :class:`NluIntentNotRecognized`)."""
    intent_filter: typing.Optional[typing.List[str]] = None
    """A list of intent names to restrict the NLU resolution on."""
    session_id: typing.Optional[str] = None
    """The id of the session, if there is an active session.

    Note
    ----

    In contrast to the Snips Hermes protocol, the session id is optional.
    """

    # ------------
    # Rhasspy only
    # ------------

    wakeword_id: typing.Optional[str] = None
    """Optional id of the wakeword used to activate the ASR.

    Note
    ----

    This is a Rhasspy-only attribute.
    """
    lang: typing.Optional[str] = None
    """Optional language of the session.

    Note
    ----

    This is a Rhasspy-only attribute.
    """

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/nlu/query"``
        """
        return "hermes/nlu/query"


@dataclass
class NluIntentParsed(Message):
    """An intent is successfully parsed.

    The NLU component returns this message as a result of the intent resolution
    requested by a :class:`NluQuery` message.

    .. admonition:: MQTT message

      Topic
        ``hermes/nlu/intentParsed``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - input
            - String
            - The user input that has generated this intent.
          * - intent
            - JSON object
            - Structured description of the intent classification.
          * - siteId
            - String
            - Site where the user interaction took place.
          * - id
            - String (optional)
            - The request identifier from the NLU query (:class:`NluQuery`), if any.
          * - slots
            - List of JSON objects (optional)
            - Structured description of the detected slots for this intent, if any.
          * - sessionId
            - String (optional)
            - Session id of the intent detection. The client code must use it to continue
              (:class:`rhasspyhermes.dialogue.DialogueContinueSession`) or end (:class:`rhasspyhermes.dialogue.DialogueEndSession`) the session.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/nlu/intentParsed'

    Note
    ----

    This is a low-level message. It preceeds the full intent message,
    :class:`NluIntent`. To detect a specific intent parsed by the NLU component,
    it is recommended to subscribe to the latter message type.
    """

    input: str
    """The user input that has generated this intent."""
    intent: Intent
    """Structured description of the intent classification."""
    site_id: str = "default"
    """Site where the user interaction took place.

    Note
    ----

    In contrast to the Snips Hermes protocol, the site id is compulsory.
    """
    id: typing.Optional[str] = None
    """The request identifier from the NLU query (:class:`NluQuery`), if any."""
    slots: typing.Optional[typing.List[Slot]] = None
    """Structured description of the detected slots for this intent, if any."""
    session_id: typing.Optional[str] = None
    """Session id of the intent detection. The client code must use it to continue
    (:class:`rhasspyhermes.dialogue.DialogueContinueSession`) or end (:class:`rhasspyhermes.dialogue.DialogueEndSession`) the session."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/nlu/intentParsed"``
        """
        return "hermes/nlu/intentParsed"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class AsrTokenTime:
    """The time when an ASR token was detected."""

    start: float
    """Start time (in seconds) of token relative to beginning of utterance."""
    end: float
    """End time (in seconds) of token relative to beginning of utterance."""


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class AsrToken:
    """A token from an automated speech recognizer."""

    value: str
    """Text value of the token."""
    confidence: float
    """Confidence score of the token, between 0 and 1 (1 being confident)."""
    range_start: int
    """The start of the range in which the token is in the original input."""
    range_end: int
    """The end of the range in which the token is in the original input."""
    time: typing.Optional[AsrTokenTime] = None
    """Structured time when this token was detected."""


@dataclass
class NluIntent(Message):
    """Recognized intent.

    This is the main Rhasspy Hermes message an intent handler should subscribe to.
    It is sent by the dialogue manager when an intent has been detected.

    It's the intent handler's responsibility to inform the dialogue manager of what it
    should do with the current session. The handler should either send a :class:`rhasspyhermes.dialogue.DialogueContinueSession`
    or a :class:`rhasspyhermes.dialogue.DialogueEndSession` message with the current session id.

    .. admonition:: MQTT message

      Topic
        ``hermes/intent/<intentName>``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - input
            - String
            - The user input that has generated this intent.
          * - intent
            - JSON object
            - Structured description of the intent classification.
          * - siteId
            - String
            - Site where the user interaction took place. Defaults to ``"default"``.
          * - id
            - String (optional)
            - The request identifier from the NLU query (:class:`NluQuery`), if any.
          * - slots
            - List of JSON objects (optional)
            - Structured description of the detected slots for this intent, if any.
          * - sessionId
            - String (optional)
            - Session id of the intent detection. The client code must use it to continue
              (:class:`rhasspyhermes.dialogue.DialogueContinueSession`) or end (:class:`rhasspyhermes.dialogue.DialogueEndSession`) the session.
          * - customData
            - String (optional)
            - Custom data provided by message that started (:class:`rhasspyhermes.dialogue.DialogueStartSession`),
              continued (:class:`rhasspyhermes.dialogue.DialogueContinueSession`) or
              ended (:class:`rhasspyhermes.dialogue.DialogueEndSession`) the session.
          * - asrTokens
            - List of list of JSON objects (optional)
            - Structured description of the tokens the ASR captured for this intent.
              The first level of lists represents each invocation of the ASR, the second
              level represents the captured tokens in that invocation.
          * - asrConfidence
            - Number (optional)
            - Speech recognizer confidence score between 0 and 1 (1 being sure).

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/intent/<intentName>'

      Replace ``<intentName>`` by the name of the intent you're interested in.
      You can use the MQTT wildcard ``#`` is you want to receive all intents.

    Example
    -------

    >>> from rhasspyhermes.nlu import NluIntent
    >>> from rhasspyhermes.intent import Intent
    >>> nlu_intent = NluIntent("what time is it", Intent(intent_name="GetTime", confidence_score=0.95))
    >>> nlu_intent.payload()
    '{"input": "what time is it", "intent": {"intentName": "GetTime", "confidenceScore": 0.95}, "siteId": "default", "id": null, "slots": null, "sessionId": null, "customData": null, "asrTokens": null, "asrConfidence": null, "rawInput": null, "wakewordId": null, "lang": null}'
    """

    TOPIC_PATTERN = re.compile(r"^hermes/intent/(.+)$")

    input: str
    """The user input that has generated this intent."""
    intent: Intent
    """Structured description of the intent classification."""
    site_id: str = "default"
    """Site where the user interaction took place."""
    id: typing.Optional[str] = None
    """Request id from the NLU query (:class:`NluQuery`), if any.

    Note
    ----

    This is a Rhasspy-only attribute."""
    slots: typing.Optional[typing.List[Slot]] = None
    """Structured description of the detected slots for this intent, if any."""
    session_id: typing.Optional[str] = None
    """Session id of the intent detection. The client code must use it to continue
    (:class:`rhasspyhermes.dialogue.DialogueContinueSession`) or end (:class:`rhasspyhermes.dialogue.DialogueEndSession`) the session.

    Note
    ----

    In contrast to the Snips Hermes protocol, the session id is optional."""
    custom_data: typing.Optional[str] = None
    """Custom data provided by message that started (:class:`rhasspyhermes.dialogue.DialogueStartSession`),
    continued (:class:`rhasspyhermes.dialogue.DialogueContinueSession`) or
    ended (:class:`rhasspyhermes.dialogue.DialogueEndSession`) the session."""
    asr_tokens: typing.Optional[typing.List[typing.List[AsrToken]]] = None
    """Structured description of the tokens the ASR captured for this intent.
    The first level of lists represents each invocation of the ASR, the second
    level represents the captured tokens in that invocation."""
    asr_confidence: typing.Optional[float] = None
    """Speech recognizer confidence score between 0 and 1 (1 being sure)."""

    # ------------
    # Rhasspy only
    # ------------

    raw_input: typing.Optional[str] = None
    """Original query input before substitutions, such as number replacement.

    Note
    ----

    This is a Rhasspy-only attribute."""
    wakeword_id: typing.Optional[str] = None
    """Id of the wake word that triggered this session.

    Note
    ----

    This is a Rhasspy-only attribute."""
    lang: typing.Optional[str] = None
    """Language of the session.

    Note
    ----

    This is a Rhasspy-only attribute."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for a message of this type with intent name ``intent_name``.

        Returns
        -------
        str
            ``"hermes/intent/{intent_name}"``

        Example
        -------

        >>> from rhasspyhermes.nlu import NluIntent
        >>> NluIntent.topic()
        'hermes/intent/#'
        >>> NluIntent.topic(intent_name="GetTime")
        'hermes/intent/GetTime'
        """
        intent_name = kwargs.get("intent_name", "#")
        return f"hermes/intent/{intent_name}"

    @classmethod
    def get_intent_name(cls, topic: str) -> str:
        """Get intent_name from a topic."""
        match = re.match(NluIntent.TOPIC_PATTERN, topic)
        assert match, "Not an intent topic"
        return match.group(1)

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template."""
        return re.match(NluIntent.TOPIC_PATTERN, topic) is not None

    def to_rhasspy_dict(self) -> typing.Dict[str, typing.Any]:
        """Convert to Rhasspy format."""
        return {
            "intent": {
                "name": self.intent.intent_name,
                "confidence": self.intent.confidence_score,
            },
            "entities": [
                {
                    "entity": s.slot_name,
                    "value": s.value.get("value"),
                    "value_details": s.value,
                    "raw_value": s.raw_value,
                    "start": s.start,
                    "end": s.end,
                    "raw_start": (s.raw_start if s.raw_start is not None else s.start),
                    "raw_end": (s.raw_end if s.raw_end is not None else s.end),
                }
                for s in self.slots or []
            ],
            "slots": {s.slot_name: s.value.get("value") for s in self.slots or []},
            "text": self.input,
            "raw_text": self.raw_input or "",
            "tokens": self.input.split(),
            "raw_tokens": (self.raw_input or self.input).split(),
            "wakeword_id": self.wakeword_id,
        }

    @classmethod
    def make_asr_tokens(cls, tokens: typing.List[typing.Any]) -> typing.List[AsrToken]:
        """Create ASR token objects from words."""
        asr_tokens: typing.List[AsrToken] = []
        start: int = 0

        for token in tokens:
            token_str = str(token)
            asr_tokens.append(
                AsrToken(
                    value=token_str,
                    confidence=1.0,
                    range_start=start,
                    range_end=(start + len(token_str)),
                )
            )

            start += len(token_str) + 1

        return asr_tokens


@dataclass
class NluIntentNotRecognized(Message):
    """Intent not recognized.

    .. admonition:: MQTT message

      Topic
        ``hermes/nlu/intentNotRecognized``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - input
            - String
            - The user input, if any, that generated this event.
          * - siteId
            - String
            - Site where the user interaction took place. Defaults to ``"default"``.
          * - id
            - String (optional)
            - The request identifier from the NLU query (:class:`NluQuery`), if any.
          * - customData
            - String (optional)
            - Custom data provided by message that started (:class:`rhasspyhermes.dialogue.DialogueStartSession`),
              continued (:class:`rhasspyhermes.dialogue.DialogueContinueSession`) or
              ended (:class:`rhasspyhermes.dialogue.DialogueEndSession`) the session.
          * - sessionId
            - String (optional)
            - Session id of the intent detection. The client code must use it to continue
              (:class:`rhasspyhermes.dialogue.DialogueContinueSession`) or end (:class:`rhasspyhermes.dialogue.DialogueEndSession`) the session.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/nlu/intentNotRecognized'
    """

    input: str
    """The input, if any, that generated this event."""
    site_id: str = "default"
    """Site where the user interaction took place.

    Note
    ----

    In contrast to the Snips Hermes protocol, the site id is compulsory.
    """
    id: typing.Optional[str] = None
    """Request id from NLU query, if any."""
    custom_data: typing.Optional[str] = None
    """Custom data provided by message that started (:class:`rhasspyhermes.dialogue.DialogueStartSession`),
    continued (:class:`rhasspyhermes.dialogue.DialogueContinueSession`) or
    ended (:class:`rhasspyhermes.dialogue.DialogueEndSession`) the session."""
    session_id: typing.Optional[str] = None
    """Session identifier of the session that generated this intent not recognized event."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/nlu/intentNotRecognized"``
        """
        return "hermes/nlu/intentNotRecognized"

    def to_rhasspy_dict(self) -> typing.Dict[str, typing.Any]:
        """Return an empty Rhasspy intent dictionary."""
        tokens = self.input.split()
        return {
            "text": self.input,
            "raw_text": self.input,
            "tokens": tokens,
            "raw_tokens": tokens,
            "intent": {"name": "", "confidence": 0.0},
            "entities": [],
            "slots": {},
        }


@dataclass
class NluError(Message):
    """This message is published by the NLU component if an error has occurred.

    .. admonition:: MQTT message

      Topic
        ``hermes/error/nlu``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - error
            - String
            - A description of the error that occurred.
          * - siteId
            - String
            - Site where the error occurred. Defaults to ``"default"``.
          * - context
            - String (optional)
            - Additional information on the context in which the error occurred.
          * - sessionId
            - String (optional)
            - Session id, if there is an active session.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/error/nlu'

    Example
    -------

    >>> from rhasspyhermes.nlu import NluError
    >>> nlu_error = NluError(error="Unexpected error")
    >>> nlu_error.topic()
    'hermes/error/nlu'
    >>> nlu_error.payload()
    '{"error": "Unexpected error", "siteId": "default", "context": null, "sessionId": null}'
    """

    error: str
    """A description of the error that occurred."""
    site_id: str = "default"
    """The id of the site where the error occurred. Defaults to ``"default"``.

    Note
    ----

    In contrast to the Snips Hermes protocol, the site id is compulsory.
    """
    context: typing.Optional[str] = None
    """Additional information on the context in which the error occurred."""
    session_id: typing.Optional[str] = None
    """The id of the session, if there is an active session."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message.

        Returns
        -------
        str
            ``"hermes/error/nlu"``
        """
        return "hermes/error/nlu"


# ----------------------------------------------------------------------------
# Rhasspy-specific Messages
# ----------------------------------------------------------------------------


@dataclass
class NluTrain(Message):
    """Request to retrain NLU from intent graph.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/nlu/<siteId>/train``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - graphPath
            - String
            - Path to the graph file.
          * - id
            - String (optional)
            - Unique id for the training request. Appended to reply topic (:class:`NluTrainSuccess`).
          * - graphFormat
            - String (optional)
            - Format of the graph file.
          * - sentences
            - Dictionary (optional)
            - TODO
          * - slots
            - Dictionary (optional)
            - TODO

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        TODO

    Note
    ----

    This is a Rhasspy-only message.
    """

    TOPIC_PATTERN = re.compile(r"^rhasspy/nlu/([^/]+)/train$")

    graph_path: str
    """Path to the graph file."""
    id: typing.Optional[str] = None
    """Unique id for the training request."""
    graph_format: typing.Optional[str] = None
    """Optional format of the graph file."""
    sentences: typing.Optional[typing.Dict[str, str]] = None
    """TODO"""
    slots: typing.Optional[typing.Dict[str, typing.List[str]]] = None
    """TODO"""

    @classmethod
    def is_site_in_topic(cls) -> bool:
        """True if site id is in topic."""
        return True

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"rhasspy/nlu/{site_id}/train"``
        """
        site_id = kwargs.get("site_id", "+")
        return f"rhasspy/nlu/{site_id}/train"

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template."""
        return re.match(NluTrain.TOPIC_PATTERN, topic) is not None

    @classmethod
    def get_site_id(cls, topic: str) -> typing.Optional[str]:
        """Get site id from a topic."""
        match = re.match(NluTrain.TOPIC_PATTERN, topic)
        assert match, "Not a train topic"
        return match.group(1)


@dataclass
class NluTrainSuccess(Message):
    """Result from successful training.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/nlu/<siteId>/trainSuccess``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - id
            - String (optional)
            - Unique id from the training request (:class:`NluTrain`).

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'rhasspy/nlu/<siteId>/trainSuccess'

    Note
    ----

    This is a Rhasspy-only message.
    """

    TOPIC_PATTERN = re.compile(r"^rhasspy/nlu/([^/]+)/trainSuccess$")

    id: typing.Optional[str] = None
    """Unique id from training request."""

    @classmethod
    def is_site_in_topic(cls) -> bool:
        """True if site id is in topic."""
        return True

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"rhasspy/nlu/{site_id}/trainSuccess"``
        """
        site_id = kwargs.get("site_id", "+")
        return f"rhasspy/nlu/{site_id}/trainSuccess"

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template"""
        return re.match(NluTrainSuccess.TOPIC_PATTERN, topic) is not None

    @classmethod
    def get_site_id(cls, topic: str) -> typing.Optional[str]:
        """Get site id from a topic."""
        match = re.match(NluTrainSuccess.TOPIC_PATTERN, topic)
        assert match, "Not a trainSuccess topic"
        return match.group(1)



"""Messages for hermes/dialogueManager"""
import typing
from collections.abc import Mapping
from dataclasses import dataclass
from enum import Enum

from dataclasses_json import DataClassJsonMixin, LetterCase, dataclass_json
from dataclasses_json.core import Json

from .base import Message


class DialogueActionType(str, Enum):
    """Type of session init objects.

    Values
    ------
    ACTION
        Use this type when you need the end user to respond

    NOTIFICATION
        Use this type when you only want to inform the user of something without
        expecting a response.
    """

    ACTION = "action"
    NOTIFICATION = "notification"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DialogueAction(DataClassJsonMixin):
    """Dialogue session action.

    Attributes
    ----------
    can_be_enqueued: bool
        If true, the session will start when there is no pending one on this
        site. Otherwise, the session is just dropped if there is running one.

    type: DialogueActionType = ACTION

    text: typing.Optional[str] = None
        Text that the TTS should say at the beginning of the session.

    intent_filter: typing.Optional[typing.List[str]] = None
        A list of intents names to restrict the NLU resolution on the first query.

    send_intent_not_recognized: bool = False
        Indicates whether the dialogue manager should handle non-recognized
        intents by itself or send them for the client to handle.
    """

    can_be_enqueued: bool
    type: DialogueActionType = DialogueActionType.ACTION
    text: typing.Optional[str] = None
    intent_filter: typing.Optional[typing.List[str]] = None
    send_intent_not_recognized: bool = False


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DialogueNotification(DataClassJsonMixin):
    """Dialogue session notification.

    Attributes
    ----------
    text: str
        Text the TTS should say

    type: DialogueActionType = NOTIFICATION
    """

    text: str
    type: DialogueActionType = DialogueActionType.NOTIFICATION


class DialogueSessionTerminationReason(str, Enum):
    """The reason why the session was ended.

    Values
    ------
    NOMINAL
        The session ended as expected (an endSession message was received)

    ABORTED_BY_USER
        The session aborted by the user

    INTENT_NOT_RECOGNIZED
        The session ended because no intent was successfully detected

    TIMEOUT
        The session timed out because no response from one of the components or
        no continue/end session message in a timely manner.

    ERROR
        The session failed with an error
    """

    NOMINAL = "nominal"
    ABORTED_BY_USER = "abortedByUser"
    INTENT_NOT_RECOGNIZED = "intentNotRecognized"
    TIMEOUT = "timeout"
    ERROR = "error"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DialogueSessionTermination:
    """Dialogue session termination type.

    Attributes
    ----------
    reason: DialogueSessionTerminationReason
        The reason why the session was ended
    """

    reason: DialogueSessionTerminationReason


# -----------------------------------------------------------------------------


@dataclass
class DialogueStartSession(Message):
    """Start a dialogue session.

    You can send this message to programmatically initiate a new session. The
    Dialogue Manager will start the session by asking the TTS to say the text
    (if any) and wait for the answer of the end user.

    Attributes
    ----------
    init: Union[DialogueAction, DialogueNotification]
        Session initialization description

    site_id: str = "default"
        Site where to start the session

    custom_data: Optional[str] = None
        Additional information that can be provided by the handler. Each message
        related to the new session - sent by the Dialogue Manager - will contain
        this data.

    lang: Optional[str] = None
        Language of the session.
    """

    init: typing.Union[DialogueAction, DialogueNotification]
    site_id: str = "default"
    custom_data: typing.Optional[str] = None
    lang: typing.Optional[str] = None

    # pylint: disable=W0221
    @classmethod
    def from_dict(
        cls: typing.Type["DialogueStartSession"],
        message_dict: Json,
        *,
        infer_missing=False
    ) -> "DialogueStartSession":
        assert isinstance(message_dict, Mapping)
        init = message_dict.pop("init")
        if init["type"] == DialogueActionType.NOTIFICATION:
            message_dict["init"] = DialogueNotification.from_dict(init)
        else:
            message_dict["init"] = DialogueAction.from_dict(init)

        return super(DialogueStartSession, cls).from_dict(
            message_dict, infer_missing=infer_missing
        )

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/dialogueManager/startSession"


@dataclass
class DialogueSessionQueued(Message):
    """Sent when a dialogue session has been queued.

    Attributes
    ----------
    session_id: str
        Session identifier that was enqueued

    site_id: str = "default"
        Site where the user interaction will take place

    custom_data: typing.Optional[str] = None
        Custom data provided in the startSession message
    """

    session_id: str
    site_id: str = "default"
    custom_data: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        return "hermes/dialogueManager/sessionQueued"


@dataclass
class DialogueSessionStarted(Message):
    """Sent when a dialogue session has been started.

    Attributes
    ----------
    session_id: str
        Session identifier that was started

    site_id: str = "default"
        Site where the user interaction is taking place

    custom_data: typing.Optional[str] = None
        Custom data provided in the startSession message

    lang: Optional[str] = None
        Language of the session.
    """

    session_id: str
    site_id: str = "default"
    custom_data: typing.Optional[str] = None
    lang: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/dialogueManager/sessionStarted"


@dataclass
class DialogueContinueSession(Message):
    """Sent when a dialogue session should be continued.

    Attributes
    ----------
    session_id: str
        Identifier of the session to continue

    site_id: str = "default"
        Site where the user interaction is taking place

    text: Optional[str] = None
        The text the TTS should say to start this additional request of the
        session.

    intent_filter: Optional[List[str]] = None
        A list of intents names to restrict the NLU resolution on the answer of
        this query.

    custom_data: Optional[str] = None
        An update to the session's custom data. If not provided, the custom data
        will stay the same.

    send_intent_not_recognized: bool = False
        Indicates whether the dialogue manager should handle non recognized
        intents by itself or send them for the client to handle.

    slot: typing.Optional[str] = None
        Unused

    lang: Optional[str] = None
        Language of the session.
        Leave empty to use setting from start of session.
    """

    session_id: str
    site_id: str = "default"
    custom_data: typing.Optional[str] = None

    text: typing.Optional[str] = None
    intent_filter: typing.Optional[typing.List[str]] = None
    send_intent_not_recognized: bool = False
    slot: typing.Optional[str] = None

    # ------------
    # Rhasspy only
    # ------------

    lang: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/dialogueManager/continueSession"


@dataclass
class DialogueEndSession(Message):
    """Sent when a dialogue session should be ended.

    Attributes
    ----------
    session_id: str
        Identifier of the session to end

    site_id: str = "default"
        Site where the user interaction is taking place

    text: Optional[str] = None
        The text the TTS should say to end the session

    custom_data: Optional[str] = None
        An update to the session's custom data. If not provided, the custom data
        will stay the same.
    """

    session_id: str
    site_id: str = "default"
    text: typing.Optional[str] = None
    custom_data: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/dialogueManager/endSession"


@dataclass
class DialogueSessionEnded(Message):
    """Sent when a dialogue session has ended.

    Attributes
    ----------
    termination: DialogueSessionTermination
        Structured description of why the session has been ended

    session_id: str
        Session identifier of the ended session

    site_id: str = "default"
        Site where the user interaction took place

    custom_data: Optional[str] = None
        Custom data provided in the start/continue/end session messages
    """

    termination: DialogueSessionTermination
    session_id: str
    site_id: str = "default"
    custom_data: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/dialogueManager/sessionEnded"


@dataclass
class DialogueIntentNotRecognized(Message):
    """Intent not recognized.

    Only sent when send_intent_not_recognized is True.

    Attributes
    ----------
    session_id: str
        Session identifier of the session that generated this intent not
        recognized event.

    site_id: str = "default"
        Site where the user interaction took place

    input: typing.Optional[str] = None
        NLU input that generated this event

    custom_data: typing.Optional[str] = None
        Custom data provided by start/continue session messages
    """

    session_id: str
    site_id: str = "default"
    input: typing.Optional[str] = None
    custom_data: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/dialogueManager/intentNotRecognized"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DialogueConfigureIntent:
    """Enable/disable specific intent in configure message.

    Attributes
    ----------
    intent_id: str
        Name of intent to enable/disable

    enable: bool
        True if intent should be enabled
    """

    intent_id: str
    enable: bool


@dataclass
class DialogueConfigure(Message):
    """Enable/disable specific intents for future dialogue sessions.

    Attributes
    ----------
    intents: List[DialogueConfigureIntent]
        List of intents and whether to enable/disable them

    site_id: str = "default"
        Id of site to configure
    """

    intents: typing.List[DialogueConfigureIntent]
    site_id: str = "default"

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/dialogueManager/configure"


# ----------------------------------------------------------------------------
# Rhasspy-Only Messages
# ----------------------------------------------------------------------------


@dataclass
class DialogueError(Message):
    """Error from dialogue manager component.

    Attributes
    ----------
    error: str
        A description of the error that occurred

    site_id: str = "default"
        The id of the site where the error occurred

    context: Optional[str] = None
        Additional information on the context in which the error occurred

    session_id: Optional[str] = None
        The id of the session, if there is an active session
    """

    error: str
    site_id: str = "default"
    context: typing.Optional[str] = None
    session_id: typing.Optional[str] = None

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type."""
        return "hermes/error/dialogueManager"



"""Command-line interface to rhasspyhermes"""
import argparse
import dataclasses
import io
import json
import logging
import os
import sys
import threading
from pathlib import Path
from uuid import uuid4

import paho.mqtt.client as mqtt

from .base import Message

_LOGGER = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


def main():
    """Main method"""
    # Parse command-line arguments
    args = get_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    _LOGGER.debug(args)

    # Connect to MQTT broker
    connected_event = threading.Event()
    client = mqtt.Client()

    def on_connect(client, userdata, flags, rc):
        try:
            # Signal main thread
            connected_event.set()
        except Exception:
            _LOGGER.exception("on_connect")

    def on_disconnect(client, userdata, flags, rc):
        try:
            # Automatically reconnect
            _LOGGER.info("Disconnected. Trying to reconnect...")
            client.reconnect()
        except Exception:
            _LOGGER.exception("on_disconnect")

    # Connect
    client.on_connect = on_connect
    client.on_disconnect = on_disconnect

    def log_subscribe(topic):
        _LOGGER.debug("Subscribed to %s", topic)
        mqtt.Client.subscribe(client, topic)

    client.subscribe = log_subscribe

    _LOGGER.debug("Connecting to %s:%s", args.host, args.port)
    client.connect(args.host, args.port)

    client.loop_start()

    try:
        site_id = "default"
        if args.site_id:
            # Use first site_id
            site_id = args.site_id[0]
            args.site_id = set(args.site_id)

        # Call sub-commmand
        args.func(args, client, site_id)
    except KeyboardInterrupt:
        pass
    finally:
        _LOGGER.debug("Shutting down")
        client.loop_stop()


def get_args() -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rhasspy_hermes", description="rhasspy_hermes"
    )
    parser.add_argument(
        "--host", default="localhost", help="MQTT host (default: localhost)"
    )
    parser.add_argument(
        "--port", type=int, default=1883, help="MQTT port (default: 1883)"
    )
    parser.add_argument(
        "--site-id",
        action="append",
        help="Hermes site_id(s) to listen for (default: all)",
    )
    parser.add_argument(
        "--print-topics",
        action="store_true",
        help="Print MQTT topics with JSON messages",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG messages to the console"
    )

    # Create subparsers for each sub-command
    sub_parsers = parser.add_subparsers()
    sub_parsers.required = True
    sub_parsers.dest = "command"

    # --------------
    # transcribe-wav
    # --------------
    transcribe_parser = sub_parsers.add_parser(
        "transcribe-wav", help="Transcribe WAV file(s)"
    )
    transcribe_parser.set_defaults(func=transcribe)
    transcribe_parser.add_argument(
        "wav_file", nargs="*", default=[], help="Path(s) to WAV file(s)"
    )
    transcribe_parser.add_argument(
        "--stdin-files",
        "-f",
        action="store_true",
        help="Read WAV file paths from stdin instead of WAV data",
    )

    # ----------------
    # recognize-intent
    # ----------------
    recognize_parser = sub_parsers.add_parser(
        "recognize-intent", help="Recognize intent(s) from text"
    )
    recognize_parser.set_defaults(func=recognize)
    recognize_parser.add_argument(
        "sentence", nargs="*", default=[], help="Sentences to recognize"
    )

    # --------------
    # speak-sentence
    # --------------
    speak_parser = sub_parsers.add_parser("speak-sentence", help="Speak sentence(s)")
    speak_parser.add_argument("sentence", nargs="*", help="Sentence(s) to speak")
    speak_parser.add_argument(
        "--language", default="", help="Language for text to speech"
    )
    speak_parser.set_defaults(func=speak)

    # ---------
    # wait-wake
    # ---------
    wake_parser = sub_parsers.add_parser(
        "wait-wake", help="Wait until wake word is detected"
    )
    wake_parser.add_argument(
        "--toggle",
        action="store_true",
        help="Toggle hotword service on/off around detection",
    )
    wake_parser.add_argument(
        "--wakeword-id",
        action="append",
        default=["default"],
        help="Wake word id(s) to wait for",
    )
    wake_parser.set_defaults(func=wake)

    return parser.parse_args()


# -----------------------------------------------------------------------------


def publish(client, message: Message, **topic_args):
    """Publish a Hermes message to MQTT."""
    try:
        _LOGGER.debug("-> %s", message)
        topic = message.topic(**topic_args)
        payload = json.dumps(dataclasses.asdict(message))
        _LOGGER.debug("Publishing %s char(s) to %s", len(payload), topic)
        client.publish(topic, payload)
    except Exception:
        _LOGGER.exception("on_message")


def check_site_id(args, json_payload):
    """Return True if site_id matches one of configured site_ids"""
    if args.site_id:
        site_id = json_payload.get("site_id", "default")
        return site_id in args.site_id

    return True


def print_json(args, topic: str, message: Message):
    """Print Hermes message as single line of JSON to stdout"""
    if args.print_topics:
        print(topic, end=" ")

    json.dump(dataclasses.asdict(message), sys.stdout)
    print("")
    sys.stdout.flush()


# -----------------------------------------------------------------------------


def transcribe(args, client, site_id):
    """Transcribe one or more WAV files using hermes/asr"""
    from .asr import AsrStartListening, AsrStopListening, AsrTextCaptured
    from .audioserver import AudioFrame

    client.subscribe(AsrTextCaptured.topic())
    frame_topic = AudioFrame.topic(site_id=site_id)

    if args.wav_file:
        # Read WAV paths from command-line arguments
        def get_wavs():
            for wav_path in args.wav_file:
                wav_path = Path(wav_path)
                with open(wav_path, "rb") as wav_file:
                    yield str(wav_path), wav_file

    elif args.stdin_files:
        # Read WAV paths from stdin (one per line)
        def get_wavs():
            for wav_path in sys.stdin:
                wav_path = Path(wav_path.strip())
                with open(wav_path, "rb") as wav_file:
                    yield str(wav_path), wav_file

    else:
        # Read WAV data from stdin
        def get_wavs():
            if os.isatty(sys.stdin.fileno()):
                print("Reading WAV data from stdin...", file=sys.stderr)

            wav_bytes = sys.stdin.buffer.read()
            with io.BytesIO(wav_bytes) as wav_file:
                yield "<stdin>", wav_file

    for wav_name, wav_file in get_wavs():
        _LOGGER.debug("Transcribing %s", wav_name)

        done_event = threading.Event()
        result_topic = ""
        text_captured = None
        session_id = str(uuid4())

        def on_message(client, userdata, msg):
            nonlocal result_topic, text_captured
            try:
                if msg.topic == AsrTextCaptured.topic():
                    # Verify site_id/session_id
                    json_payload = json.loads(msg.payload)
                    if check_site_id(args, json_payload) and (
                        json_payload.get("session_id", "") == session_id
                    ):
                        # Matched
                        result_topic = msg.topic
                        text_captured = AsrTextCaptured(**json_payload)
                        done_event.set()
            except Exception:
                _LOGGER.exception("transcribe.on_message")

        client.on_message = on_message

        with wav_file:
            # startListening
            publish(client, AsrStartListening(site_id=site_id, session_id=session_id))

            # Send WAV chunks (audioFrame)
            for wav_chunk in AudioFrame.iter_wav_chunked(
                wav_file, frames_per_chunk=2048
            ):
                client.publish(frame_topic, wav_chunk)

            # stopListening
            publish(client, AsrStopListening(site_id=site_id, session_id=session_id))

        _LOGGER.debug(
            "Waiting for textCaptured (%s, session_id=%s)", wav_name, session_id
        )

        # Wait for textCaptured
        done_event.wait()

        # Print result
        assert text_captured is not None
        print_json(args, result_topic, text_captured)


# -----------------------------------------------------------------------------


def recognize(args, client, site_id):
    """Recognize intent(s) from one or more sentences using hermes/nlu"""
    from .nlu import NluQuery, NluIntent, NluIntentNotRecognized, NluError

    client.subscribe(NluIntent.topic(intent_name="#"))
    client.subscribe(NluIntentNotRecognized.topic())
    client.subscribe(NluError.topic())

    if args.sentence:
        # Read sentences from arguments
        sentences = args.sentence
    else:
        # Read sentences from stdin
        if os.isatty(sys.stdin.fileno()):
            print("Reading sentences from stdin...", file=sys.stderr)

        sentences = args.stdin

    # Process sentences
    for sentence in sentences:
        sentence = sentence.strip()
        _LOGGER.debug("Processing '%s'", sentence)

        done_event = threading.Event()
        result_topic = ""
        result_message = None
        queryId = str(uuid4())
        session_id = str(uuid4())

        def on_message(client, userdata, msg):
            nonlocal result_topic, result_message
            try:
                if NluIntent.is_topic(msg.topic):
                    # Verify site_id/id/session_id
                    json_payload = json.loads(msg.payload)
                    if (
                        check_site_id(args, json_payload)
                        and (json_payload.get("session_id", "") == session_id)
                        and (json_payload.get("id", "") == queryId)
                    ):
                        # Matched
                        result_topic = msg.topic
                        result_message = NluIntent(**json_payload)
                        done_event.set()
                elif msg.topic == NluIntentNotRecognized.topic():
                    # Verify site_id/id/session_id
                    json_payload = json.loads(msg.payload)
                    if (
                        check_site_id(args, json_payload)
                        and (json_payload.get("id", "") == queryId)
                        and (json_payload.get("session_id", "") == session_id)
                    ):
                        # Not recognized
                        result_topic = msg.topic
                        result_message = NluIntentNotRecognized(**json_payload)
                        done_event.set()
                elif msg.topic == NluError.topic():
                    # Verify site_id/session_id
                    json_payload = json.loads(msg.payload)
                    if check_site_id(args, json_payload) and (
                        json_payload.get("session_id", "") == session_id
                    ):
                        # Error
                        result_topic = msg.topic
                        result_message = NluError(**json_payload)
                        _LOGGER.error(result_message.error)
                        done_event.set()
            except Exception:
                _LOGGER.exception("recognize.on_message")

        client.on_message = on_message

        # Send query
        publish(
            client,
            NluQuery(
                input=sentence, site_id=site_id, id=queryId, session_id=session_id
            ),
        )

        # Wait for response
        _LOGGER.debug("Waiting for intent/notRecognized/error (id=%s)", queryId)
        done_event.wait()

        # Print result
        assert result_message is not None
        print_json(args, result_topic, result_message)


# -----------------------------------------------------------------------------


def speak(args, client, site_id):
    """Speak one or more sentences using hermes/tts"""
    from .tts import TtsSay, TtsSayFinished

    client.subscribe(TtsSayFinished.topic())

    if args.sentence:
        # Read sentences from arguments
        sentences = args.sentence
    else:
        # Read sentences from stdin
        if os.isatty(sys.stdin.fileno()):
            print("Reading sentences from stdin...", file=sys.stderr)

        sentences = args.stdin

    # Speak sentences
    for sentence in sentences:
        sentence = sentence.strip()
        _LOGGER.debug("Speaking '%s'", sentence)

        done_event = threading.Event()
        result_topic = ""
        result_message = None
        sayId = str(uuid4())
        session_id = str(uuid4())

        def on_message(client, userdata, msg):
            nonlocal result_topic, result_message
            try:
                if msg.topic == TtsSayFinished.topic():
                    # Verify site_id/id/session_id
                    json_payload = json.loads(msg.payload)
                    if (
                        check_site_id(args, json_payload)
                        and (json_payload.get("session_id", "") == session_id)
                        and (json_payload.get("id", "") == sayId)
                    ):
                        # Matched
                        result_topic = msg.topic
                        result_message = TtsSayFinished(**json_payload)
                        done_event.set()
            except Exception:
                _LOGGER.exception("speak.on_message")

        client.on_message = on_message

        # Send say
        publish(
            client,
            TtsSay(
                text=sentence,
                lang=args.language,
                site_id=site_id,
                id=sayId,
                session_id=session_id,
            ),
        )

        # Wait for finished
        _LOGGER.debug("Waiting for finished (id=%s)", sayId)
        done_event.wait()

        # Print result
        assert result_message is not None
        print_json(args, result_topic, result_message)


# -----------------------------------------------------------------------------


def wake(args, client, site_id):
    """Wait until wake word is detected"""
    from .wake import HotwordToggleOn, HotwordDetected, HotwordToggleOff

    assert args.wakeword_id, "No wake word ids"
    for wakeword_id in args.wakeword_id:
        client.subscribe(HotwordDetected.topic(wakeword_id=wakeword_id))

    done_event = threading.Event()
    result_topic = ""
    result_message = None

    def on_message(client, userdata, msg):
        nonlocal result_topic, result_message
        try:
            if HotwordDetected.is_topic(msg.topic):
                # Verify wakeword_id
                wakeword_id = HotwordDetected.get_wakeword_id(msg.topic)
                if wakeword_id in args.wakeword_id:
                    # Matched
                    json_payload = json.loads(msg.payload)
                    result_topic = msg.topic
                    result_message = HotwordDetected(**json_payload)
                    done_event.set()
        except Exception:
            _LOGGER.exception("speak.on_message")

    client.on_message = on_message

    if args.toggle:
        # toggleOn
        publish(client, HotwordToggleOn(site_id=site_id))

    # Wait for detection
    _LOGGER.debug("Waiting for detected (wakeword_id=%s)", args.wakeword_id)
    done_event.wait()

    if args.toggle:
        # toggleOff
        publish(client, HotwordToggleOff(site_id=site_id))

    # Print result
    assert result_message is not None
    print_json(args, result_topic, result_message)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    main()
